from .attention import GlobalSelfAttention, CausalSelfAttention, CrossAttention

from .autoencoder import Encoder1D, Encoder2D, Encoder3D
from .autoencoder import Decoder1D, Decoder2D, Decoder3D

from .tsmixer import TSMixerBlock



import tensorflow as tf
from typing import Tuple
from functools import partial
from math import pi


def _get_complex_tensor_from_tuple(x):
    """
    Get a complex tensor from a tuple of two real-valued tensors.

    Parameters
    ----------
    x : tf.Tensor | list | tuple
        Input tensor, either `tf.Tensor` (real part) or tuple/list (real- and imaginary part).
        If only the real part is provided, the imaginary part is assumed 0.

    Returns
    -------
    y : tf.Tensor
        Tensor with `dtype=tf.complex`.

    Raises
    ------
    ValueError
        If shapes of real- and imaginary part to not match.

    """

    if not isinstance(x, (tuple, list)) or len(x) != 2:
        real = x
        imag = tf.zeros_like(x)
    else:
        real, imag = x
    # Check shapes.
    if real.shape != imag.shape:
        raise ValueError(
            "Input `x` should be a tuple of two tensors - real and imaginary."
            "Both the real and imaginary parts should have the same shape. "
            f"Received: x[0].shape = {real.shape}, x[1].shape = {imag.shape}"
        )

    complex_input = tf.complex(real=real, imag=imag)
    return complex_input


# === FFT ===
def _fft(x: tuple, fn: callable) -> Tuple[tf.Tensor, tf.Tensor]:
    complex_input = _get_complex_tensor_from_tuple(x)
    complex_output = fn(complex_input)
    return tf.math.real(complex_output), tf.math.imag(complex_output)


# === real valued FFT ===
def _rfft(x: tf.Tensor, fn: callable) -> Tuple[tf.Tensor, tf.Tensor]:
    complex_output = fn(x)
    return tf.math.real(complex_output), tf.math.imag(complex_output)


def _irfft(x: tf.Tensor, fn: callable, n: tuple = None) -> tf.Tensor:
    complex_input = _get_complex_tensor_from_tuple(x)
    complex_output = fn(complex_input, fft_length=n)
    return tf.math.real(complex_output), tf.math.imag(complex_output)


# === FFT derivatives ===
def fftfreq(n, d, rad=False):
    fs = 1.0 / d
    df = fs / tf.cast(n, tf.float32)
    fft_freqs = tf.cast(tf.range(n) - n // 2, tf.float32)
    fft_freqs *= df

    if rad:
        fft_freqs *= (2.0 * pi)

    fft_freqs = tf.roll(fft_freqs, shift=n // 2, axis=0)
    return fft_freqs


# === FFT derivative ===
def fft_derivative(x, axis=-1, d=1.0, n=1):
    """
    Derivative in Fourier space

    Parameters
    ----------
    x : tf.Tensor
        Input array, can be either real or complex valued.
    axis : int (optional)
        Axis along which `x` is derived.
        Defaults to `-1`.
    d : float (optional)
        Discretization of `x` along the dimension of `axis`.
        Defaults to `1.0`.
    n : int (optional)
        Order of differentiation.
        Defaults to `1`.

    Returns
    -------
    dx : tf.Tensor
        nth derivative of `x` along `axis`.

    Examples
    --------
    >>> n = 32
    >>> t = tf.range(0, 2 * pi, 2 * pi / 32)
    >>> dt = 2 * pi / 32
    >>> x = tf.sin(t)
    >>> dxdt_ref = tf.cos(t)
    >>> dxdt = fft_derivative(x, d=dt)
    >>> tf.reduce_max(dxdt_ref - dxdt)  
    tf.Tensor(1.0728836e-06, shape=(), dtype=float32)

    Notes
    -----
    For higher orders, the derivatives may get noisy.
    The intensity depends on the signal length, i.e.,
    this effect is less significant for short signals.

    """
    
    nx = tf.shape(x)[axis]
    dim = len(tf.shape(x))

    # tf.signal.fft is always along last axis --> move axis to last dimension
    if dim > 2:
        perm = list(range(dim))
        perm.append(perm.pop(axis))

        x = tf.transpose(x, perm=perm)

    x = _get_complex_tensor_from_tuple(x)
    freqs = fftfreq(n=nx, d=d, rad=True)

    complex_diff = tf.signal.ifft(tf.pow((1j * tf.cast(freqs, x.dtype)), tf.cast(n, x.dtype)) * tf.signal.fft(x))
    diff = tf.math.real(complex_diff)

    # revert permutation
    if dim > 2:
        perm = list(range(dim))
        perm.insert(axis + 1, perm.pop())
        diff = tf.transpose(diff, perm=perm)

    return diff


# === derived functions
def fft_fn(x):
    return partial(_fft, fn=tf.signal.fft)(x)


def fft2_fn(x):
    return partial(_fft, fn=tf.signal.fft2d)(x)


def fft3_fn(x):
    return partial(_fft, fn=tf.signal.fft3d)(x)


def ifft_fn(x):
    return partial(_fft, fn=tf.signal.ifft)(x)


def ifft2_fn(x):
    return partial(_fft, fn=tf.signal.ifft2d)(x)


def ifft3_fn(x):
    return partial(_fft, fn=tf.signal.ifft3d)(x)


def rfft_fn(x):
    return partial(_rfft, fn=tf.signal.rfft)(x)


def rfft2_fn(x):
    return partial(_rfft, fn=tf.signal.rfft2d)(x)


def rfft3_fn(x):
    return partial(_rfft, fn=tf.signal.rfft3d)(x)


def irfft_fn(x, n=None):
    y_real, _ = partial(_irfft, fn=tf.signal.irfft, n=n)(x)
    return y_real


def irfft2_fn(x, n=None):
    y_real, _ = partial(_irfft, fn=tf.signal.irfft2d, n=n)(x)
    return y_real


def irfft3_fn(x, n=None):
    y_real, _ = partial(_irfft, fn=tf.signal.irfft3d, n=n)(x)
    return y_real



from keras import layers
from keras import models
from keras import saving
from ...blocks.tsmixer import TSMixerBlock
from ...ops.helper import _IterableVars


@saving.register_keras_serializable(package="Kerex.Models", name="TSMixer")
class TSMixer(models.Model, _IterableVars):
    """
    TSMixer model, cf. [Chen et al.](https://doi.org/10.48550/arXiv.2303.06053)

    Parameters
    ----------
    sequence_length : int
        Length of the predicted output sequence.
    num_hidden : int | list | tuple, optional
        Number of units in dense layer.
        Determines the number of TSMixer residual blocks (1 layer if `num_hidden` of type `int` or `None`).
        If `None`, the feature space is expanded by factor of 4 for the respective layer.
        Defaults to `None`.
    norm : str | list | tuple, optional {`"LN"`, `"BN"`}
        Normalization type, can be layer normalization (`"LN"`) or batch normalization (`"BN"`).
        Defaults to `"LN"`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation for dense layers.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    dropout_rate : float | list | tuple, optional
        Dropout rate.
        Defaults to 0.1.
    name : str, optional
        Name of the layer.
        If `None`, `name` is automatically inherited from the class name `"TSMixerBlock"`.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `keras.layers.Layer` super-class.

    Notes
    -----
    Adapted for Keras3 from https://github.com/google-research/google-research/tree/master/tsmixer

    Examples
    --------
    For `None` arguments in `num_hidden`, the respective projection dimension depends on the number of features in the input.
    Explicit arguments of type `int` result in the respective projection dimension.
    The following code creates a TSMixer model with three residual blocks.
    >>> from keras import ops
    >>> x = ops.ones((1, 256, 32))  # time series of length 256 with 32 features
    >>> model = TSMixer(sequence_length=512, num_hidden=[None, 32, 16])
    >>> model.build(input_shape=x.shape)
    >>> y = model(x)
    >>> y.shape
    TensorShape([1, 512, 32])

    """

    def __init__(
        self,
        sequence_length,
        num_hidden=None,
        norm="LN",
        activation="relu",
        dropout_rate=0.1,
        name=None,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.sequence_length = sequence_length

        # residual block arguments
        self.set_vars(num_hidden=num_hidden, norm=norm, dropout_rate=dropout_rate)
        self.activation = activation

        self.residual_blocks = models.Sequential([
            TSMixerBlock(
                num_hidden=n,
                norm=nrm,
                activation=self.activation,
                dropout_rate=d
            ) for n, nrm, d in zip(
                self.num_hidden,
                self.norm,
                self.dropout_rate
            )
        ], name="tsmixer_blocks")

        self.output_section = models.Sequential([
            layers.Permute(dims=(2, 1), name="transpose_1"),
            layers.Dense(units=self.sequence_length, name="dense"),
            layers.Permute(dims=(2, 1), name="transpose_2")
        ], name="output_stage")

    def build(self, input_shape):
        if self.built:
            return
        
        self.residual_blocks.build(input_shape=input_shape)
        forward_shape = self.residual_blocks.compute_output_shape(input_shape=input_shape)

        self.output_section.build(input_shape=forward_shape)

        self.built = True

    def call(self, inputs):
        x = self.residual_blocks(inputs)
        x = self.output_section(x)
        return x
    
    def compute_output_shape(self, input_shape):
        output_shape = self.residual_blocks.compute_output_shape(input_shape)
        output_shape = self.output_section.compute_output_shape(output_shape)

        return output_shape
    
    def get_config(self):
        config = super().get_config()
        config.update({
            "sequence_length": self.sequence_length,
            "num_hidden": self.num_hidden,
            "norm": self.norm,
            "activation": saving.serialize_keras_object(self.activation),
            "dropout_rate": self.dropout_rate
        })
        return config
    
    @classmethod
    def from_config(cls, config):
        activation_cfg = config.pop("activation")
        config.update({"activation": saving.deserialize_keras_object(activation_cfg)})

        return cls(**config)
    
    def summary(self, *args, **kwargs):
        """
        Prints the summary of the model.

        Here `super().summary()` is called with `expand_nested=True`

        Parameters
        ----------
        line_length: int, optional
            Total length of printed lines
            (e.g. set this to adapt the display to different
            terminal window sizes).
        positions: list, optional
            Relative or absolute positions of log elements
            in each line. If not provided, becomes
            `[0.3, 0.6, 0.70, 1.]`. Defaults to `None`.
        print_fn: callable, optional
            Print function to use. By default, prints to `stdout`.
            If `stdout` doesn't work in your environment, change to `print`.
            It will be called on each line of the summary.
            You can set it to a custom function
            in order to capture the string summary.
        show_trainable: bool, optional
            Whether to show if a layer is trainable.
            Defaults to `False`.
        layer_range: list | tuple, optional
            a list or tuple of 2 strings,
            which is the starting layer name and ending layer name
            (both inclusive) indicating the range of layers to be printed
            in summary. It also accepts regex patterns instead of exact
            names. In this case, the start predicate will be
            the first element that matches `layer_range[0]`
            and the end predicate will be the last element
            that matches `layer_range[1]`.
            By default `None` considers all layers of the model.

        Raises
        ------
        ValueError
            If `summary()` is called before the model is built.
        
        """

        kwargs.pop("expand_nested", None)
        return super().summary(expand_nested=True, *args, **kwargs)
    


from keras import saving
from keras import layers
from keras import Sequential
from keras import ops
from keras.src.layers.convolutional.base_conv import BaseConv
from keras.src.layers.convolutional.base_separable_conv import BaseSeparableConv
from keras.src.layers.convolutional.base_conv_transpose import BaseConvTranspose
from keras.src import activations
from ...ops import get_layer
from importlib import import_module


@saving.register_keras_serializable(package="Kerex.Layers", name="SpatialAttentionModule")
class SpatialAttentionModule(layers.Layer):
    """
    Spatial Attention Module (SAM)

    Parameters
    ----------
    rank : int, optional
        Rank of the convolutions.
        Defaults to 2.
    kernel_size : int | tuple, optional
        An integer or a tuple/list of integers indicating the kernel size of the convolutional projection layer.
        Defaults to 5.
    activation : str, optional
        Activation function for the convolutional layer.
        Defaults to `"sigmoid"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format of the convolutions.
        Can be either `"channels_first"` or `"channels_last"`.
        Defaults to `"channels_last"`.
    name : str, optional
        Name of the layer. Defaults to `"spatial_attention_module"`.
    **kwargs : Additional keyword arguments for the `Layer` super-class.

    Notes
    -----
    This is not a wrapper but a CAM layer.
    May be used on it's own, but here it serves as a building block for the 
    [Convolutional Block Attention Module (CBAM)](https://arxiv.org/abs/1807.06521).
    
    More information on the SAM can be found here: https://www.digitalocean.com/community/tutorials/attention-mechanisms-in-computer-vision-cbam#channel-attention-module-cam
    
    """

    def __init__(self, rank=2, kernel_size=5, activation="sigmoid", data_format="channels_last", name="spatial_attention_module", **kwargs):
        super().__init__(name=name, **kwargs)
        self.rank = rank
        self.kernel_size = kernel_size
        self.activation = activation
        self.data_format = data_format
        self.channel_axis = -1 if self.data_format == "channels_last" else 1

        self.projection = get_layer(f"Conv{self.rank}D", filters=1, kernel_size=self.kernel_size, padding="same", data_format=self.data_format)
        self.activation_fn = activations.get(self.activation)
    
    def build(self, input_shape):
        """
        Build method of `SpatialAttentionModule`
        
        Builds all sub-modules and sets `self.built=True`.

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        """

        if self.built:
            return
        
        input_shape_projection = list(input_shape)

        input_shape_projection[self.channel_axis] = 2

        self.projection.build(input_shape=tuple(input_shape_projection))
        
        self.built = True

    def call(self, inputs):
        """
        Call method of `SpatialAttentionModule`

        Parameters
        ----------
        inputs : KerasTensor
            Input tensor for `SpatialAttentionModule`.

        Returns
        -------
        outputs : KerasTensor
            Refined features.

        """

        channel_avg = ops.average(inputs, axis=self.channel_axis)
        channel_max = ops.max(inputs, axis=self.channel_axis)
        x = ops.stack([channel_avg, channel_max], axis=self.channel_axis)  # shape (batch, h, w, 2)
        x = self.projection(x)  # shape (batch, h, w, 1)
        x = self.activation_fn(x)  # shape (batch, h, w, 1)

        return x  # shape (batch, h, w, 1)
    
    def compute_output_shape(self, input_shape):
        """
        Compute output shape

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        Returns
        -------
        output_shape : tuple
            Shape of the output of `SpatialAttentionModule`.

        """
        
        output_shape = list(input_shape)
        output_shape[self.channel_axis] = 1
        
        return tuple(output_shape)
    
    def get_config(self):
        """
        Get config method.
        Required for serialization.

        Returns
        -------
        config : dict
            Dictionary with the configuration of `SpatialAttentionModule`

        """
        
        config: dict = super().get_config()
        config.update({
            "rank": self.rank,
            "kernel_size": self.kernel_size,
            "activation": self.activation,
            "data_format": self.data_format
        })
        return config


@saving.register_keras_serializable(package="Kerex.Layers.Wrapper", name="CBAM")
class CBAM(layers.Wrapper):
    """
    Convolutional Block Attention Module Wrapper

    Can be used to wrap any convolutional Keras layer.
    The layer refines the latent space variable / features by sequentially applying
    - channel attention from ChannelAttentionModule, followed by
    - spatial attention from SpatialAttentionModule.

    Parameters
    ----------
    layer : Layer
        Keras `layer` or `Sequential` model to be wrapped.
        If this is a `Sequential` model, all layers have to be convolutional.
    reduction_ratio : int, optional
        Size of the bottleneck in relation to the input size of the shared MLP in `ChannelAttentionModule`.
        Defaults to 8.
    kernel_size : int | tuple, optional
        An integer or a tuple/list of integers indicating the kernel size of the convolutional projection layer.
        This is parameter for the `SpatialAttentionModule`.
        Defaults to 5.
    channel_activation : str, optional
        Activation in `ChannelAttentionModule`.
        Defaults to `"sigmoid"`.
    spatial_activation : str, optional
        Activation in `SpatialAttentionModule`.
        Defaults to `"sigmoid"`.
    name : str, optional
        Name of the wrapper, that is prepended to the name of the wrapped layer.
        Defaults to `"cbam"`.
    **kwargs : Additional keyword arguments for the `Wrapper` super-class.

    Raises
    ------
    ValueError
        If `layer` is not a Convolutional layer.

    Examples
    --------
    >>> from keras import layers, ops
    >>> x = ops.ones((2, 32, 32, 3))
    >>> conv_layer = layers.Conv2D(filters=8, kernel_size=3)
    >>> cbam_conv_layer = CBAM(conv_layer)
    >>> layer_output = conv_layer(x)
    >>> cbam_output = cbam_conv_layer(x)
    >>> layer_output.shape == cbam_output.shape
    True
    >>> ops.convert_to_numpy(ops.equal(cbam_output, layer_output)).any()
    np.False_

    """

    def __init__(self, layer, reduction_ratio=8, kernel_size=5, channel_activation="sigmoid", spatial_activation="sigmoid", name="cbam", **kwargs):
        if issubclass(type(layer), Sequential):
            if not all([isinstance(type(l), (BaseConv, BaseConvTranspose, BaseSeparableConv)) for l in layer.layers]):
                raise ValueError(f"CBAM is only compatible with convolutional layers, received a Sequential model with {[l.__class__ for l in layer.layers]} layers")
        elif issubclass(type(layer), (BaseConv, BaseConvTranspose, BaseSeparableConv)):
            pass
        else:
            raise ValueError(f"CBAM is only compatible with convolutional layers, received {layer.__class__}")
        
        name = "_".join([layer.name, name])
        super().__init__(layer=layer, name=name, **kwargs)
        self.reduction_ratio = reduction_ratio
        self.kernel_size = kernel_size
        self.channel_activation = channel_activation
        self.spatial_activation = spatial_activation

        self.channel_attention = getattr(import_module(name="...layers.pooling", package=__package__), f"ChannelAttentionModule{self.layer.rank}D")(
            reduction_ratio=self.reduction_ratio,
            activation=self.channel_activation,
            data_format=self.layer.data_format,
            keepdims=True
        )
        self.spatial_attention = SpatialAttentionModule(
            rank=self.layer.rank,
            kernel_size=self.kernel_size,
            activation=self.spatial_activation,
            data_format=self.layer.data_format
        )

    def build(self, input_shape):
        """
        Build method of `CBAM`
        
        Builds all sub-modules and sets `self.built=True`.

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        """

        super().build(input_shape=input_shape)
        output_shape = self.layer.compute_output_shape(input_shape=input_shape)
        self.channel_attention.build(input_shape=output_shape)
        self.spatial_attention.build(input_shape=output_shape)

    def call(self, inputs):
        """
        Call method of `CBAM`.
        This layer applies (1) channel-wise and (2) spatial attention to refine the features.

        Parameters
        ----------
        inputs : KerasTensor
            Input tensor for `CBAM`.

        Returns
        -------
        outputs : KerasTensor
            Refined features.

        """

        x = self.layer(inputs)

        # apply attention sequentially
        x = ops.multiply(x, self.channel_attention(x))
        x = ops.multiply(x, self.spatial_attention(x))
        return x
    
    def get_config(self):
        """
        Get config method.
        Required for serialization.

        Returns
        -------
        config : dict
            Dictionary with the configuration of `CBAM`.

        """

        config: dict = super().get_config()
        config.update({
            "reduction_ratio": self.reduction_ratio,
            "kernel_size": self.kernel_size,
            "channel_activation": self.channel_activation,
            "spatial_activation": self.spatial_activation
        })
        return config
    
    def compute_output_shape(self, input_shape):
        """
        Compute output shape

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        Returns
        -------
        output_shape : tuple
            Shape of the output of `CBAM`.

        Notes
        -----
        `CBAM` does not modify the output shape.
        Hence, this layer simply calls the `compute_output_shape` method of the wrapped layer.

        """

        return self.layer.compute_output_shape(input_shape)
    


from .base_fno import BaseFNO
from keras import saving


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="FNO1D")
class FNO1D(BaseFNO):
    def __init__(
        self, 
        filters, 
        modes, 
        activation="gelu", 
        merge_layer="add", 
        data_format="channels_last", 
        use_bias=True, 
        kernel_initializer="he_normal", 
        bias_initializer="zeros", 
        kernel_constraint=None, 
        bias_constraint=None, 
        kernel_regularizer=None, 
        bias_regularizer=None, 
        name=None, 
        **kwargs
    ):
        super().__init__(
            rank=1, 
            filters=filters, 
            modes=modes, 
            activation=activation, 
            merge_layer=merge_layer,  
            data_format=data_format, 
            use_bias=use_bias, 
            kernel_initializer=kernel_initializer, 
            bias_initializer=bias_initializer, 
            kernel_constraint=kernel_constraint, 
            bias_constraint=bias_constraint, 
            kernel_regularizer=kernel_regularizer, 
            bias_regularizer=bias_regularizer, 
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="FNO2D")
class FNO2D(BaseFNO):
    def __init__(
        self, 
        filters, 
        modes, 
        activation="gelu", 
        merge_layer="add", 
        data_format="channels_last", 
        use_bias=True, 
        kernel_initializer="he_normal", 
        bias_initializer="zeros", 
        kernel_constraint=None, 
        bias_constraint=None, 
        kernel_regularizer=None, 
        bias_regularizer=None, 
        name=None, 
        **kwargs
    ):
        super().__init__(
            rank=2, 
            filters=filters, 
            modes=modes, 
            activation=activation, 
            merge_layer=merge_layer,  
            data_format=data_format, 
            use_bias=use_bias, 
            kernel_initializer=kernel_initializer, 
            bias_initializer=bias_initializer, 
            kernel_constraint=kernel_constraint, 
            bias_constraint=bias_constraint, 
            kernel_regularizer=kernel_regularizer, 
            bias_regularizer=bias_regularizer, 
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="FNO3D")
class FNO3D(BaseFNO):
    def __init__(
        self, 
        filters, 
        modes, 
        activation="gelu", 
        merge_layer="add", 
        data_format="channels_last", 
        use_bias=True, 
        kernel_initializer="he_normal", 
        bias_initializer="zeros", 
        kernel_constraint=None, 
        bias_constraint=None, 
        kernel_regularizer=None, 
        bias_regularizer=None, 
        name=None, 
        **kwargs
    ):
        super().__init__(
            rank=3, 
            filters=filters, 
            modes=modes, 
            activation=activation, 
            merge_layer=merge_layer,  
            data_format=data_format, 
            use_bias=use_bias, 
            kernel_initializer=kernel_initializer, 
            bias_initializer=bias_initializer, 
            kernel_constraint=kernel_constraint, 
            bias_constraint=bias_constraint, 
            kernel_regularizer=kernel_regularizer, 
            bias_regularizer=bias_regularizer, 
            name=name, 
            **kwargs
        )
        


from functools import partial
from tensorflow.python.framework import ops
from tensorflow.python.ops import random_ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import cond as tf_cond
from tensorflow.python.ops.image_ops_impl import _AssertAtLeast3DImage
import tensorflow as tf
from math import pi


""" RANDOM ROTATION DATA AUGMENTATION """
AFFINE_TRANSFORM_INTERPOLATIONS = (
    "nearest",
    "bilinear",
)
AFFINE_TRANSFORM_FILL_MODES = (
    "constant",
    "nearest",
    "wrap",
    "reflect",
)


def _deg_to_rad(deg: float):
    return deg * pi / 180


def affine_transform(
    image,
    transform,
    interpolation="bilinear",
    fill_mode="constant",
    fill_value=0,
    data_format="channels_last",
):
    if interpolation not in AFFINE_TRANSFORM_INTERPOLATIONS:
        raise ValueError(
            "Invalid value for argument `interpolation`. Expected of one "
            f"{AFFINE_TRANSFORM_INTERPOLATIONS}. Received: "
            f"interpolation={interpolation}"
        )
    if fill_mode not in AFFINE_TRANSFORM_FILL_MODES:
        raise ValueError(
            "Invalid value for argument `fill_mode`. Expected of one "
            f"{AFFINE_TRANSFORM_FILL_MODES}. Received: fill_mode={fill_mode}"
        )
    if len(image.shape) not in (3, 4):
        raise ValueError(
            "Invalid image rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"image.shape={image.shape}"
        )
    if len(transform.shape) not in (1, 2):
        raise ValueError(
            "Invalid transform rank: expected rank 1 (single transform) "
            "or rank 2 (batch of transforms). Received input with shape: "
            f"transform.shape={transform.shape}"
        )
    # unbatched case
    need_squeeze = False
    if len(image.shape) == 3:
        image = tf.expand_dims(image, axis=0)
        need_squeeze = True
    if len(transform.shape) == 1:
        transform = tf.expand_dims(transform, axis=0)

    if data_format == "channels_first":
        image = tf.transpose(image, (0, 2, 3, 1))

    affined = tf.raw_ops.ImageProjectiveTransformV3(
        images=image,
        transforms=tf.cast(transform, dtype=tf.float32),
        output_shape=tf.shape(image)[1:-1],
        fill_value=fill_value,
        interpolation=interpolation.upper(),
        fill_mode=fill_mode.upper(),
    )
    affined = tf.ensure_shape(affined, image.shape)

    if data_format == "channels_first":
        affined = tf.transpose(affined, (0, 3, 1, 2))
    if need_squeeze:
        affined = tf.squeeze(affined, axis=0)
    return affined


def _get_rotation_matrix(image: tf.Tensor, angle: int | float, rank: int):
    shape = array_ops.shape(image)
    if rank == 4:
        b = shape[0]
        h = shape[1]
        w = shape[2]
    
    elif rank == 3:
        b = 1
        h = shape[0]
        w = shape[1]
    else:
        raise ValueError(f"Random rotation only supports batched (4D) or unbatched (3D) tensors!")
    
    cos_theta = tf.math.cos(angle)
    sin_theta = tf.math.sin(angle)
    image_height = tf.cast(h, dtype=cos_theta.dtype)
    image_width = tf.cast(w, dtype=cos_theta.dtype)

    x_offset = (
        (image_width - 1) - (cos_theta * (image_width - 1) - sin_theta * (image_height - 1))
    ) / 2.0
    
    y_offset = (
        (image_height - 1) - (sin_theta * (image_width - 1) + cos_theta * (image_height - 1))
    ) / 2.0

    outputs = tf.concat([
        tf.math.cos(angle)[:, tf.newaxis],
        -tf.math.sin(angle)[:, tf.newaxis],
        x_offset[:, tf.newaxis],
        tf.math.sin(angle)[:, tf.newaxis],
        tf.math.cos(angle)[:, tf.newaxis],
        y_offset[:, tf.newaxis],
        tf.zeros((b, 2))
    ], axis=1)

    if rank == 3:
        outputs = tf.squeeze(outputs, axis=0)
    return outputs


def _random_rotation(image: tf.Tensor, random_func: callable, scope_name: str):
    with ops.name_scope(None, scope_name, [image]) as scope:
        image = ops.convert_to_tensor(image, name="image")
        image = _AssertAtLeast3DImage(image)
        shape = image.get_shape()

        def fn_rank3():
            # rotation last dimension
            angle = random_func(shape=[1, ])

            rotation_matrix = _get_rotation_matrix(image, angle, rank=3)
            rotated_image = affine_transform(
                image=image,
                transform=rotation_matrix,
                interpolation="bilinear",
                fill_mode="constant",
                fill_value=0.0,
                data_format="channels_last"
            )
            return rotated_image

        def fn_rank4():
            # rotation along last dimension, too!
            batch_size = array_ops.shape(image)[0]
            angle = random_func(shape=[batch_size, ])

            rotation_matrix = _get_rotation_matrix(image, angle, rank=4)
            rotated_image = affine_transform(
                image=image,
                transform=rotation_matrix,
                interpolation="bilinear",
                fill_mode="constant",
                fill_value=0.0,
                data_format="channels_last"
            )
            return rotated_image

        if shape.ndims is None:
            rank = array_ops.rank(image)
            return tf_cond.cond(math_ops.equal(rank, 3), fn_rank3, fn_rank4)
        if shape.ndims == 3:
            return fn_rank3()
        elif shape.ndims == 4:
            return fn_rank4()
        else:
            raise ValueError(
                '\'image\' (shape %s) must have either 3 or 4 dimensions.' % shape)


def random_rotation(image, angle, seed=None):
    """
    Apply random rotation to an image or 2-D tensor.
    Can be savely used within `tf.data.Dataset.map()`.

    Parameters
    ----------
    image : tf.Tensor
        image or 2-D tensor to rotate
    angle : int | tuple | list
        (min_angle, max_angle) rotation angle in degree for random rotation. A single value results in `angle=(-angle, angle)`.
    seed : int
        Random seed, defaults to `None`.

    Returns
    -------
    image : tf.Tensor
        Randomly rotated `image`.

    Raises
    ------
    ValueError
        If more than 2 values for `angle` are provided.
    ValueError
        If unsupported type for `angle` is provided.

    """

    if angle is None or angle == 0.0:
        return image
    
    if isinstance(angle, (int, float)):
        # only allow positive rotation
        minval = min(0.0, angle)
        maxval = max(0.0, angle)
    elif isinstance(angle, (list, tuple)):
        if len(angle) > 2:
            raise ValueError(f"Too many values for rotation angle, received angle={angle}, expected 2 values.")
        if len(angle) == 2:
            minval = min(angle)
            maxval = max(angle)
        else:
            # only allow positive rotation
            minval = min(0.0, *angle)
            maxval = max(0.0, *angle)
    else:
        raise ValueError(f"Unsupported type for angle {type(angle)}. Expected (int, float, tuple, list)")
    
    random_func = partial(random_ops.random_uniform, minval=_deg_to_rad(minval), maxval=_deg_to_rad(maxval), seed=seed)
    return _random_rotation(image=image, random_func=random_func, scope_name="random_rotation")



import jax
import jax.numpy as jnp
from typing import Tuple
from functools import partial


def _get_complex_tensor_from_tuple(x) -> jax.lax.complex:
    """
    Get a complex tensor from a tuple of two real-valued tensors.

    Parameters
    ----------
    x : jax.lax.real | list | tuple
        Input tensor, either `jax.lax.real` (real part) or tuple/list (real- and imaginary part).
        If only the real part is provided, the imaginary part is assumed 0.

    Returns
    -------
    y : jax.lax.complex
        Tensor with `dtype=jax.lax.complex`.

    Raises
    ------
    ValueError
        If shapes of real- and imaginary part to not match.
    ValueError
        If at least on of the inputs is not of type float.

    """

    if not isinstance(x, (tuple, list)) or len(x) != 2:
        real = x
        imag = jnp.zeros_like(x)
    else:
        real, imag = x
    # Check shapes.
    if real.shape != imag.shape:
        raise ValueError(
            "Input `x` should be a tuple of two tensors - real and imaginary."
            "Both the real and imaginary parts should have the same shape. "
            f"Received: x[0].shape = {real.shape}, x[1].shape = {imag.shape}"
        )
    # Ensure dtype is float.
    if not jnp.issubdtype(real.dtype, jnp.floating) or not jnp.issubdtype(
        imag.dtype, jnp.floating
    ):
        raise ValueError(
            "At least one tensor in input `x` is not of type float."
            f"Received: x={x}."
        )
    
    complex_input = jax.lax.complex(real, imag)
    return complex_input


# === FFT ===
def _fft(x: tuple, fn: callable, **kwargs) -> Tuple[jnp.ndarray, jnp.ndarray]:
    complex_input = _get_complex_tensor_from_tuple(x)
    complex_output = fn(complex_input, **kwargs)
    return jnp.real(complex_output), jnp.imag(complex_output)


# === real valued FFT ===
def _rfft(x: jnp.ndarray, fn: callable, **kwargs) -> Tuple[jnp.ndarray, jnp.ndarray]:
    complex_output = fn(x, **kwargs)
    return jnp.real(complex_output), jnp.imag(complex_output)


def _irfft(x: jnp.ndarray, fn: callable, **kwargs) -> jnp.ndarray:
    complex_input = _get_complex_tensor_from_tuple(x)
    complex_output = fn(complex_input, **kwargs)
    return jnp.real(complex_output), jnp.imag(complex_output)


# === FFT derivative ===
def fft_derivative(x, axis=-1, d=1.0, n=1):
    """
    Derivative in Fourier space

    Parameters
    ----------
    x : array_like
        Input array, can be either real or complex valued.
    axis : int (optional)
        Axis along which `x` is derived.
        Defaults to `-1`.
    d : float (optional)
        Discretization of `x` along the dimension of `axis`.
        Defaults to `1.0`.
    n : int (optional)
        Order of differentiation.
        Defaults to `1`.

    Returns
    -------
    dx : array_like
        nth derivative of `x` along `axis`.

    Examples
    --------
    >>> t = jnp.linspace(0, 2 * jnp.pi, 32, endpoint=False)
    >>> dt = jnp.mean(jnp.diff(t))
    >>> x = jnp.sin(t)
    >>> dx_ref = jnp.cos(t)
    >>> dx = fft_derivative(x, d=dt, n=1.0)
    >>> jnp.allclose(dx_ref, dx, atol=1e-5)
    True

    Notes
    -----
    For higher orders, the derivatives may get noisy.
    The intensity depends on the signal length, i.e.,
    this effect is less significant for short signals.

    """
    
    nx = x.shape[axis]

    x = _get_complex_tensor_from_tuple(x)
    freqs = jnp.fft.fftfreq(n=nx, d=d) * 2 * jnp.pi

    complex_diff = jnp.fft.ifft((1j * freqs)**n * jnp.fft.fft(x, axis=axis), axis=axis)
    diff = jnp.real(complex_diff)

    return diff


# === derived functions ===
def fft_fn(x):
    return partial(_fft, fn=jnp.fft.fft)(x)


def fft2_fn(x):
    return partial(_fft, fn=jnp.fft.fft2)(x)


def fft3_fn(x):
    return partial(_fft, fn=jnp.fft.fftn, axes=(-3, -2, -1))(x)


def ifft_fn(x):
    return partial(_fft, fn=jnp.fft.ifft)(x)


def ifft2_fn(x):
    return partial(_fft, fn=jnp.fft.ifft2)(x)


def ifft3_fn(x):
    return partial(_fft, fn=jnp.fft.ifftn, axes=(-3, -2, -1))(x)


def rfft_fn(x):
    return partial(_rfft, fn=jnp.fft.rfft)(x)


def rfft2_fn(x):
    return partial(_rfft, fn=jnp.fft.rfft2)(x)


def rfft3_fn(x):
    return partial(_rfft, fn=jnp.fft.rfftn, axes=(-3, -2, -1))(x)


def irfft_fn(x, n=None):
    if isinstance(n, tuple):
        n, = n  # unpack tuple
    y_real, _ = partial(_irfft, fn=jnp.fft.irfft, n=n)(x)
    return y_real


def irfft2_fn(x, n=None):
    y_real, _ = partial(_irfft, fn=jnp.fft.irfft2, s=n)(x)
    return y_real


def irfft3_fn(x, n=None):
    y_real, _ = partial(_irfft, fn=jnp.fft.irfftn, s=n, axes=(-3, -2, -1))(x)
    return y_real


### FFT derivatives ###



import tensorflow as tf
import numpy as np
import json
from pathlib import Path
import warnings
from string import ascii_lowercase


class TFRecordsHandlerBase:
    """
    Base dataloader class for TFRecord files

    This base class can be subclassed to easily generate input pipelines for ML 
    based on [`tf.data.Dataset`](https://www.tensorflow.org/api_docs/python/tf/data/Dataset).

    This base class can
    - automatically write numpy arrays to tfrecord files,
    - generate a basic `tf.data.Dataset` from a list of tfrecord files
    
    Parameters
    ----------
    path : str | Path, otptional
        Path to where TFRecord files (and metadata.json) are stored.

    Examples
    --------
    Let's write some arbitrary data (`x` and `y`) to `data.tfrecords`. `x` is of type `float32`, `y` is of type `int16`.
    >>> import numpy as np
    >>> x = np.arange(32).reshape(4, 8).astype(np.float32)
    >>> y = np.array([1, 0, 0, 1]).astype(np.int16)
    >>> dataloader = TFRecordsHandlerBase(path=".")
    >>> dataloader.setup(x=x, y=y)
    >>> dataloader.metadata
    {'x': {'dtype': 'float32', 'shape': (4, 8), 'shape_names': ('a', 'b')}, 'y': {'dtype': 'int16', 'shape': (4,), 'shape_names': ('a',)}}
    >>> dataloader.write_tfrecords(file="data.tfrecords", x=x, y=y)
    >>> dataset = dataloader.get_dataset(files=["data.tfrecords"])
    >>> x_tf, y_tf = next(iter(dataset.take(1)))
    >>> x_tf
    <tf.Tensor: shape=(32,), dtype=float32, numpy=
    array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9., 10., 11., 12.,
        13., 14., 15., 16., 17., 18., 19., 20., 21., 22., 23., 24., 25.,
        26., 27., 28., 29., 30., 31.], dtype=float32)>
    >>> y_tf
    <tf.Tensor: shape=(4,), dtype=int16, numpy=array([1, 0, 0, 1], dtype=int16)>

    """

    def __init__(self, path=None):
        if path:
            self.path = Path(path)
            try:
                self.metadata = self.read_metadata()
            except FileNotFoundError:
                pass

        self.valid_dtype = (np.number, np.ndarray)

    def setup(self, **kwargs):
        """
        Initialize dimensions dictionary by passing all features to the setup

        Builds a metadata file from named numpy arrays.
        The metadata.json contains
        - the dtype (e.g., `"float32"`),
        - the shape (e.g., `(5,2)`), and
        - the shape_names (e.g., `(a, b)`)
        for each key (named numpy input file.)

        Parameters
        ----------
        **kwargs : Keyword arguments
            Have to be numpy arrays. The key determines the key in the metadata.json

        Returns
        -------
        metadata : dict
            A dictionary containing essential information about the arrays.

        Raises
        ------
        TypeError
            If not all inputy are numpy arrays.

        Examples
        --------
        >>> dataloader.setup(x=np.ones((5, 2)), y=np.ones((1)))
        >>> dataloader.metadata
        {'x': {'dtype': 'float64', 'shape': (5,2), 'shape_names': ('a','b')}, 'y': {'dtype': 'float64', 'shape': (1,), 'shape_names': ('a',)}}

        """

        if all(isinstance(v, self.valid_dtype) for v in kwargs.values()):
            self.metadata = {
                k: {
                    "dtype": str(v.dtype),
                    "shape": v.shape,
                    "shape_names": tuple([d for _, d in zip(v.shape, ascii_lowercase)])
                    } for k, v in kwargs.items()
                }
        else:
            raise TypeError(f"Unknown data type, received {[type(item) for item in kwargs.values()]}")
        
    @property
    def get_metadata(self):
        """
        Returns the self.metadata

        Returns
        -------
        metadata : dict
            The metadata or "architecture" of the TFRecord files.

        """

        try:
            return self.metadata
        except AttributeError as e:
            warnings.warn("dataloader.metadata is not set yet. Run dataloader.setup() to initialize it", RuntimeWarning)
    
    def write_metadata(self, path=None):
        """
        Write self.metadata to json

        Parameters
        ----------
        path : str | Path, optional
            Path to save metadata.json to.
            If `self.path` is defined, `path` is ignored. Else, `self.path` is set to `path`.
            Defaults to `None`.

        Raises
        ------
        ValueError
            If `path=None` and `self.path` is not set.

        """

        if hasattr(self, "path"):
            path = self.path
        elif path is not None:
            setattr(self, "path", path)
            print(f"Set self.path={self.path}")
        else:
            raise ValueError(f"Please define path when initializing or here, received None.")
        
        with open(Path(self.path, "tfrecord_metadata.json"), "w") as fp:
            json.dump(self.metadata, fp)

    def read_metadata(self):
        """
        Read metadata.json

        Returns
        -------
        metadata : dict
            The metadata (architecture) of the TFRecord files.

        """
        with open(Path(self.path, 'tfrecord_metadata.json'), 'r') as fp:
            metadata = json.load(fp)

        # cast lists to tuples to restore original design of "metadata"
        for v in metadata.values():
            for k_, v_ in v.items():
                if isinstance(v_, list):
                    v.update({k_: tuple(v_)})

        return metadata
    
    def get_dim_keys(self, key: str) -> list:
        """
        Get dimension-names of a `key` in `self.metadata`

        Parameters
        ----------
        key : str
            Key of the entry to obtain the dimension-names from 

        Returns
        -------
        dim : list
            A sorted list with all dimension-names of `key`.

        Raises
        ------
        KeyError
            If key does not exist in `self.metadata`

        """

        try:
            return sorted([f"{key}_{d}" for d in self.metadata[key]["shape_names"]])
        except KeyError as e:
            raise KeyError(f"Key '{key}' does not exist in metadata.") from e
    
    
    def get_dtype(self, key):
        """
        Get dtype of a `key` in `self.metadata`

        Parameters
        ----------
        key : str
            Key of the entry to obtain the dtype from.

        Raises
        ------
        KeyError
            If key does not exist in `self.metadata`

        """

        try:
            return self.metadata[key]["dtype"]
        except KeyError as e:
            raise KeyError(f"Key '{key}' does not exist in metadata.") from e
    
    def _get_data_structure(self, **kwargs) -> dict:
        """
        Returns a dictionary with the structure of the TFRecord file

        Parameters
        ----------
        **kwargs : Additional optional keyword arguments

        Returns
        -------
        d : dict
            A dictionary holding the features and the respective dimensions.
        
        Raises
        ------
        RuntimeError
            If called without any `**kwargs` and `self.metadata` is not set yet.
        TypeError
            If `**kwargs` contains other dtypes than `numpy.ndarray`
        
        Notes
        -----
        This function behaves differently depending on the optional `**kwargs` argument.
        - If `**kwargs` contains `numpy.ndarrays`, the function serializes the features and their dimensions to store them in TFRecords.
        - If no `**kwargs` are defined, the function returns the features and their dimensions to instances of `tf.io.FixedLenFeature`.

        The latter requires the existance of a `self.metadata` dictionary, which defines the architecture of the TFRecord files.

        **This function is called internally, it should not be necessary to call it manually!**

        """
        if not kwargs:
            # """
            # Example usage
            #     return dictionary with keys and tf-specific data types, e.g.,
            #     >>> raw_feature = dict(feature_name=tf.io.FixedLenFeature([], tf.string))
            #     >>> dimensions = dict(x=tf.io.FixedLenFeature([], tf.int64), y=tf.io.FixedLenFeature([], tf.int64))
            #     >>> return {**raw_feature, **dimensions}
            # """
            if not hasattr(self, "metadata"):
                raise RuntimeError(f"'self.metadata' is not yet initialized. Run 'self.setup()' with all features once to set it up.")
            
            features = {k: tf.io.FixedLenFeature([], tf.string) for k in self.metadata.keys()}
            dimensions = {f"{k}_{s}": tf.io.FixedLenFeature([], tf.int64) for k in self.metadata.keys() for s in self.metadata[k]["shape_names"]}
            return {**features, **dimensions}

        if all(isinstance(v, self.valid_dtype) for v in kwargs.values()):
            # """
            # Example usage
            #     >>> serialized_feature = dict(feature_name=self._bytes_feature(tf.io.serialize_tensor(feature)))
            #     >>> dimensions = {k: self._int64_feature(feature.shape[i]) for i, k in enumerate(self.metadata))}
            #     >>> return {**serialied_feature, **dimensions}
            # """
            if not hasattr(self, 'metadata'):
                self.setup(**kwargs)

            features = {k: self._bytes_feature(tf.io.serialize_tensor(v)) for k, v in kwargs.items()}
            dimensions = {f"{k}_{d}": self._int64_feature(v.shape[i]) for k, v in kwargs.items() for i, d in enumerate(self.metadata[k]["shape_names"])}
            return {**features, **dimensions}
        
        raise TypeError(f"Unknown data type.")
        
    def _parse_write(self, **kwargs):
        """
        Writes data in `**kwargs` to `tf.train.Example`.
        
        Parameters
        ----------
        **kwargs : Additional optional keyword arguments
            Has to contain `numpy.ndarrays`.

        Returns
        -------
        y : tf.train.Example
            `x` encoded in TFRecord data format.

        Notes
        -----
        Key names have to resemble the feature names,
        e.g., `x=np.ones((32,))` writes the serialized array `np.ones((32,))` with key `x` to the `tf.train.Example`

        """

        if not kwargs:
            return
        
        if not hasattr(self, "metadata"):
            self.setup(**kwargs)
            
        return tf.train.Example(features=tf.train.Features(feature=self._get_data_structure(**kwargs)))
    
    def _parse_read(self, serialized_data):
        """
        Reads serialized `tf.train.Example` and parses them to `tf.Tensor`.

        Parameters
        ----------
        serialized_data : tf.train.Example
            Serialized data as stored in TFRecord files.

        Returns
        -------
        data : tf.Tensor
            Decoded data from TFRecord files.

        """

        structure = tf.io.parse_single_example(serialized_data, self._get_data_structure())

        if len(self.metadata.keys()) == 1:
            k = list(self.metadata.keys())[0]
            dim_keys = self.get_dim_keys(key=k)
            dtype = self.get_dtype(key=k)

            serialized_data = structure.pop(k)
            shape = tuple([structure.pop(dk) for dk in dim_keys])
            return tf.reshape(tf.io.parse_tensor(serialized_data, out_type=dtype), shape=shape)
        
        features = []
        for k in self.metadata.keys():
            dim_keys = self.get_dim_keys(key=k)
            dtype = self.get_dtype(key=k)

            serialized_data = structure.pop(k)
            shape = tuple([structure.pop(dk) for dk in dim_keys])
            features.append(tf.reshape(tf.io.parse_tensor(serialized_data, out_type=dtype), shape=shape))

        return tuple(features)

    def write_tfrecords(self, file, verbose=True, **kwargs):
        """
        Write an arbitrary amount of `numpy.ndarray` to `file`.
        The file contains a serialized version of the `numpy.ndarray`.

        Parameters
        ----------
        file : str | Path
            Filehandle `(path/to/file.tfrecords)` to where the tfrecord data should be written
        verbose : bool, optional
            If `True` verbosity is enabled.
        **kwargs : Additional optional keywords for `self._parse_write`

        Notes
        -----
        See `Examples` of the class for a MWE on how to use this function.

        """
        
        file = Path(file).absolute()
        if file.exists():
            return
        
        with tf.device("/cpu:0"):
            with tf.io.TFRecordWriter(str(file)) as writer:
                out = self._parse_write(**kwargs)
                writer.write(out.SerializeToString())
            
            if verbose:
                print(f"Successfully written data to {file}")

    def get_dataset(self, files):
        """
        Get a dataset which just decodes TFRecord files

        This method can be overwritten for more complex datasets including excessive data augmentation etc.

        Parameters
        ----------
        files : list
            List of files that form the dataset.

        Returns
        -------
        dataset : tf.data.DatasetV2
            Dataset from `files`.

        Raises
        ------
        RuntimeError
            If files is an empty list.

        """
        files = [str(file) for file in files]

        if files:
            with tf.device("/cpu:0"):
                dataset = tf.data.TFRecordDataset(filenames=files)
                dataset = dataset.map(self._parse_read, num_parallel_calls=tf.data.AUTOTUNE)

                return dataset
            
        raise RuntimeError(f"Empty dataset. Please ensure that `*.tfrecords` exist.")
    
    @staticmethod
    def _bytes_feature(value):
        """
        Returns a `tf.train.BytesList` from a string/byte.

        Parameters
        ----------
        value : tf.Tensor | ndarray
            Data in either tf.Tensor or numpy ndarray.

        Returns
        -------
        feature : tf.train.Feature
            Data encoded as `tf.train.BytesList`.

        """

        if isinstance(value, tf.Tensor):
            value = value.numpy()
        return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))
    
    @staticmethod
    def _float_feature(value):
        """
        Returns a `tf.train.FloatList` from a float / double.
        
        Parameters
        ----------
        value : tf.Tensor | ndarray
            Data in either tf.Tensor or numpy ndarray.

        Returns
        -------
        feature : tf.train.Feature
            Data encoded as `tf.train.FloatList`.

        """

        return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))

    @staticmethod
    def _int64_feature(value):
        """
        Returns an `tf.train.Int64List` from a bool/enum/int/uint.

        Parameters
        ----------
        value : tf.Tensor | ndarray
            Data in either tf.Tensor or numpy ndarray.

        Returns
        -------
        feature : tf.train.Feature
            Data encoded as `tf.train.Int64List`.

        """

        if isinstance(value, tf.Tensor):
            value = value.numpy()
        return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))
        


from keras import models
from keras import regularizers, initializers, constraints
from keras import saving
from ...ops.helper import _IterableVars
from importlib import import_module


class BaseNeuralOperator(models.Model, _IterableVars):
    """
    BaseNeuralOperator, cf. [Zongyi etl al.](https://doi.org/10.48550/arXiv.2010.08895)

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of `BaseNeuralOperator`. Must be within {1, 2, 3}.
    filters : int | list | tuple
        Number of filters for the sequential FNO layers.
        If this is a list/tuple, the length of this argument determines the number of FNO layers.
    modes : int | list | tuple
        Number of Fourier modes for FNO layers.
        An `int` results in global `modes`, a `list` allows to define the `modes` per FNO layer.
        For `rank>1`, the `modes` can be defined in terms of tuples, 
        where each entry determines the modes in the respective direction,
        e.g., `modes=(8, 4)` will result in 8 modes in y, and 4 modes in x-direction.
    input_projection_dimension : int, (optional)
        Projection dimension for the input layer.
        If `None`, there is no projection layer.
        Defaults to `None`.
    output_projection_dimension : int, (optional)
        Projection dimension for the output layer.
        If `None`, there is no projection layer.
        Defaults to `None`.
    data_format : str, (optional) {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    merge_layer : str | keras.layers.Layer, (optional) {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Merge operation in FNO layers to combine the result from the spectral convolution with the result from the bypass convolution.
        Defaults to `"add"`.
    activation : str | keras.activations.Activation | keras.layers.Layer, (optional)
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"gelu"`.
    use_bias : bool, (optional)
        If `True`, all layers use a bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, (optional)
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, (optional)
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, (optional)
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, (optional)
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, (optional)
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, (optional)
        Bias constraint.
        Defaults to `None`.
    name : str, (optional)
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"BaseNeuralOperator"`.
        Defaults to `None`.

    """

    def __init__(
        self,
        rank,
        filters,
        modes,
        input_projection_dimension=None,
        output_projection_dimension=None,
        data_format="channels_last",
        merge_layer="add",
        activation="gelu",
        use_bias=True,
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.rank = rank
        self.data_format = data_format
        self.activation = activation
        self.use_bias = use_bias
        self.merge_layer = merge_layer

        self.input_projection_dimension = input_projection_dimension
        self.output_projection_dimension = output_projection_dimension
        
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint

        # set iterable class variables for the model
        self.set_vars(
            filters=filters, 
            modes=modes
        )

        # define layers
        layer_kwargs = dict(
            data_format=self.data_format, 
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint
        )

        self.layers_ = []
        
        # add input projection
        if input_projection_dimension is not None:
            input_projection = getattr(import_module(name="...layers.conv", package=__package__), f"Conv{self.rank}D")(
                filters=self.input_projection_dimension,
                kernel_size=1, 
                name="input_projection",
                **layer_kwargs
            )
            self.layers_.append(input_projection)

        # add FNO layers
        self.layers_.extend([
            getattr(import_module(name="...layers", package=__package__), f"FNO{self.rank}D")(
                filters=f, 
                modes=m,
                activation=self.activation,
                merge_layer=self.merge_layer,
                **layer_kwargs
            ) for f, m in zip(self.filters, self.modes)
        ])

        # add output projection
        if self.output_projection_dimension is not None:
            output_projection = getattr(import_module(name="...layers.conv", package=__package__), f"Conv{self.rank}D")(
                filters=self.output_projection_dimension,  # or 1,  # this is updated in build method IF self.output_projection_filters is None
                kernel_size=1, 
                name="output_projection",
                **layer_kwargs
            )
            self.layers_.append(output_projection)

    def build(self, input_shape):
        """
        Build method of `BaseNeuralOperator`

        Builds all sub-modules and sets `self.built=True`.
        
        Parameters
        ----------
        input_shape : tuple
            Input shape

        """

        if self.built:
            return

        for layer in self.layers_:
            layer.build(input_shape=input_shape)
            input_shape = layer.compute_output_shape(input_shape=input_shape)

        self.built = True

    def call(self, inputs):
        """
        Call method of `BaseNeuralOperator`

        The `inputs` are first projected to `self.input_projection_filters`,
        then processed by sequential FNO layers,
        and finally projected to `self.output_projection_filters`.
        
        Parameters
        ----------
        inputs : KerasTensor
            Input tensor.

        Returns
        -------
        outputs : KerasTensor
            outputs of `BaseNeuralOperator`

        """
        x = inputs
        for layer in self.layers:
            x = layer(x)

        return x
    
    def compute_output_shape(self, input_shape):
        """
        Compute output shape of `BaseNeuralOperator`

        Parameters
        ----------
        input_shape : tuple
            Input shape.

        Returns
        -------
        output_shape : tuple
            Output shape.

        """

        for layer in self.layers:
            input_shape = layer.compute_output_shape(input_shape=input_shape)

        return input_shape
    
    def get_config(self):
        """
        Necessary for Keras serialization

        Returns
        -------
        config : dict
            Dictionary with the layer configuration.
            
        """

        config = super().get_config()
        config.update({
            "filters": self.filters,
            "modes": self.modes,
            "input_projection_dimension": self.input_projection_dimension,
            "output_projection_dimension": self.output_projection_dimension,
            "data_format": self.data_format,
            "merge_layer": saving.serialize_keras_object(self.merge_layer),
            "activation": saving.serialize_keras_object(self.activation),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint)
        })
        
        return config
    
    @classmethod
    def from_config(cls, config):
        """
        Necessary for Keras deserialization

        Parameters
        ----------
        cls : BaseNeuralOperator
            The `BaseNeuralOperator` class.
        config : dict
            Dictionary with the layer configuration.

        Returns
        -------
        cls : BaseNeuralOperator
            Instance of `BaseNeuralOperator` from `config`.
            
        """

        activation_cfg = config.pop("activation")
        merge_layer_cfg = config.pop("merge_layer")
        kernel_initializer_cfg = config.pop("kernel_initializer")
        bias_initializer_cfg = config.pop("bias_initializer")
        kernel_regularizer_cfg = config.pop("kernel_regularizer")
        bias_regularizer_cfg = config.pop("bias_regularizer")
        kernel_constraints_cfg = config.pop("kernel_constraint")
        bias_constraints_cfg = config.pop("bias_constraint")

        # now update with deserialized version
        config.update({
            "activation": saving.deserialize_keras_object(activation_cfg),
            "merge_layer": saving.deserialize_keras_object(merge_layer_cfg),
            "kernel_initializer": initializers.deserialize(kernel_initializer_cfg),
            "bias_initializer": initializers.deserialize(bias_initializer_cfg),
            "kernel_regularizer": regularizers.deserialize(kernel_regularizer_cfg),
            "bias_regularizer": regularizers.deserialize(bias_regularizer_cfg),
            "kernel_constraint": constraints.deserialize(kernel_constraints_cfg),
            "bias_constraint": constraints.deserialize(bias_constraints_cfg)
        })

        return cls(**config)
    
    def summary(self, *args, **kwargs):
        """
        Prints the summary of the model.

        Here `super().summary()` is called with `expand_nested=True`

        Parameters
        ----------
        line_length: int, (optional)
            Total length of printed lines
            (e.g. set this to adapt the display to different
            terminal window sizes).
        positions: list, (optional)
            Relative or absolute positions of log elements
            in each line. If not provided, becomes
            `[0.3, 0.6, 0.70, 1.]`. Defaults to `None`.
        print_fn: callable, (optional)
            Print function to use. By default, prints to `stdout`.
            If `stdout` doesn't work in your environment, change to `print`.
            It will be called on each line of the summary.
            You can set it to a custom function
            in order to capture the string summary.
        show_trainable: bool, (optional)
            Whether to show if a layer is trainable.
            Defaults to `False`.
        layer_range: list | tuple, (optional)
            a list or tuple of 2 strings,
            which is the starting layer name and ending layer name
            (both inclusive) indicating the range of layers to be printed
            in summary. It also accepts regex patterns instead of exact
            names. In this case, the start predicate will be
            the first element that matches `layer_range[0]`
            and the end predicate will be the last element
            that matches `layer_range[1]`.
            By default `None` considers all layers of the model.

        Raises
        ------
        ValueError
            If `summary()` is called before the model is built.
        
        """

        kwargs.pop("expand_nested", None)
        return super().summary(expand_nested=True, *args, **kwargs)



from keras import layers
from keras import saving
from keras.src.layers.merging.base_merge import Merge
from ...ops import get_layer


@saving.register_keras_serializable(package="Kerex.Layers.Wrapper", name="Residual")
class Residual(layers.Wrapper):
    """
    Residual Wrapper

    Can be used to add a residual connection to any Keras layer.

    Parameters
    ----------
    layer : Layer
        Keras `layer` or `Model` to be wrapped.
    merge_layer : str | Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Identifier or layer of a valid Keras merge layer, cf. https://keras.io/api/layers/merging_layers/.
        Defaults to `"add"`.
    residual_layer : str | Layer, optional
        Identifier or Keras layer that is placed in the residual connection.
        Defaults to `None`.
    name : str, optional
        Name of the wrapper, that is prepended to the name of the wrapped layer.
        Defaults to `"residual"`.
    **kwargs : Additional keyword arguments for the `Wrapper` super-class.

    Raises
    ------
    ValueError
        If `merge_layer` is not a valid Keras merge layer, cf. https://keras.io/api/layers/merging_layers/

    Notes
    -----
    This wrapper allows to implement more complex data flow within the Keras `Sequential` model.
    
    When the wrapped layer modifies the shape of the data, the `residual_layer` has to account for that,
    especially when using a `merge_layer` that expects two inputs of the same shape (e.g., `layers.Add()`).

    Examples
    --------
    The number of filters in the wrapped `Conv1D` layer matches the number of channels in the input,
    such that no `residual_layer` is necessary to merge the information using `merge_layer="add"`.
    >>> from keras import Sequential, layers, ops
    >>> x = ops.ones((1, 32, 3))
    >>> model = Sequential([Residual(layers.Conv1D(filters=3, kernel_size=3, padding="same"))])
    >>> model.build(input_shape=x.shape)
    >>> y = model(x)
    >>> y.shape
    TensorShape([1, 32, 3])
    
    The number of filters in the wrapped `Conv1D` layer **does not match** the number of channels in the input
    Hence, we have to account for the shape modification by projecting the residual information 
    by using a 1x1-`Conv1D` layer with matching number of filters.
    >>> from keras import Sequential, layers, ops
    >>> x = ops.ones((1, 32, 3))
    >>> model = Sequential([Residual(layers.Conv1D(filters=8, kernel_size=3, padding="same"), residual_layer=layers.Conv1D(filters=8, kernel_size=1))])
    >>> model.build(input_shape=x.shape)
    >>> y = model(x)
    >>> y.shape
    TensorShape([1, 32, 8])

    Another way to account for a shape missmatch between then wrapped layer output and the residual information
    is by using a `merge_layer="concatenate"`. The layer just appends the two Tensors along the channel dimension.
    >>> from keras import Sequential, layers, ops
    >>> x = ops.ones((1, 32, 3))
    >>> model = Sequential([Residual(layers.Conv1D(filters=8, kernel_size=3, padding="same"), merge_layer="concatenate")])
    >>> model.build(input_shape=x.shape)
    >>> y = model(x)
    >>> y.shape
    TensorShape([1, 32, 11])  # NOTE 8 + 3 = 11 output channels!

    """

    def __init__(self, layer, merge_layer="add", residual_layer=None, name="residual", **kwargs):
        # get channel axis
        try:
            self.data_format = layer.data_format
        except AttributeError:
            self.data_format = "channels_last"  # this is valid for dense layer for example! TODO: check if it really is ;-D

        # get merge layer
        merge_axis = -1 if self.data_format == "channels_last" else 1
        try:
            self.merge_layer = get_layer(merge_layer, axis=merge_axis)
        except TypeError:
            self.merge_layer = get_layer(merge_layer)

        if not issubclass(type(self.merge_layer), Merge):
            raise ValueError(
                f"Merge-layer {self.merge_layer} supplied to Wrapper isn't "
                "a supported merge-layer."
            )
        
        # get residual layer
        self.residual_layer = get_layer(residual_layer)
        if self.residual_layer is not None:
            if not issubclass(type(self.residual_layer), layers.Layer):
                raise ValueError(
                    f"Residual-layer {self.residual_layer} supplied to Wrapper isn't "
                    "a supported layer type. Please "
                    "ensure residual-layer is a valid Keras layer."
                )
        
        name = '_'.join([name, layer.name])
        super().__init__(layer=layer, name=name, **kwargs)

    def build(self, input_shape):
        """
        Build method of `Residual`.

        Builds all sub-modules and sets `self.built=True`.

        Parameters
        ----------
        input_shape : tuple
            Shape of the input data.

        """

        super().build(input_shape=input_shape)
        forward_shape = self.layer.compute_output_shape(input_shape)

        if self.residual_layer is not None:
            self.residual_layer.build(input_shape)
            input_shape = self.residual_layer.compute_output_shape(input_shape)

        try:
            self.merge_layer.build([forward_shape, input_shape])
        except ValueError:
            raise ValueError(f"Merge layer could not be build. Please provide a residual layer which makes the wrapper layers output and the inputs compatible in shape!")

    def call(self, inputs):
        """
        Call method of `Residual`

        Parameters
        ----------
        inputs : KerasTensor
            Input tensor for the wrapped Keras `layer` or `Model`.

        Returns
        -------
        outputs : KerasTensor
            Output tensor of the wrapped Keras `layer` or `Model`.

        """
        
        x1 = self.layer(inputs)

        if self.residual_layer is not None:
            inputs = self.residual_layer(inputs)
        return self.merge_layer([x1, inputs])
    
    def get_config(self):
        """
        Get config method.
        Required for serialization.

        Returns
        -------
        config : dict
            Dictionary with the configuration of `Residual`.

        """

        config = super().get_config()
        config.update({
            "merge_layer": saving.serialize_keras_object(self.merge_layer),
            "residual_layer": saving.serialize_keras_object(self.residual_layer)
        })
        return config
    
    @classmethod
    def from_config(cls, config):
        """
        From config method.
        Required for deserialization.

        Parameters
        ----------
        config : dict
            Dictionary with the configuration of `Residual`.

        Returns
        -------
        cls : Residual
            A instance of class `Residual` based on `config`.

        """

        layer_config = config.pop("layer")
        merge_layer_config = config.pop("merge_layer")
        residual_layer_config = config.pop("residual_layer")

        layer = saving.deserialize_keras_object(layer_config)
        merge_layer = saving.deserialize_keras_object(merge_layer_config)
        residual_layer = saving.deserialize_keras_object(residual_layer_config)

        return cls(layer=layer, merge_layer=merge_layer, residual_layer=residual_layer, **config)

    def compute_output_shape(self, input_shape):
        """
        Compute output shape

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        Returns
        -------
        output_shape : tuple
            Shape of the output of `Residual`.

        """

        forward_shape = self.layer.compute_output_shape(input_shape)

        if self.residual_layer is not None:
            input_shape = self.residual_layer.compute_output_shape(input_shape)

        output_shape = self.merge_layer.compute_output_shape([forward_shape, input_shape])

        return output_shape



from keras import saving
from .base_models import BaseFCN


@saving.register_keras_serializable(package="Kerex.Models.AutoEncoder", name="FCN1D")
class FCN1D(BaseFCN):
    """
    1-D implementation of Fully Convolutional Network (FCN)

    The model is build from a convolutional encoder- and a decoder path,
    and is able to extract features on different scale of the data
    by bisecting the features with each level of the model.

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    padding : str, optional {`"same"`, `"causal"`}
        Padding for all convolutional layers in the model.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    use_skip_connection : bool | list, optional
        Whether to use skip connections (basically transform FCN to a Unet).
        Can be defined per layer, i.e., 
        `use_skip_connection=[True, False, False]` results in a skip connection on the highest level only.
        Defaults to `False`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the (optional) information from skip connections.
        For this parameter to have an impact, there has to be at least one skip connection in the model.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"FCN1D"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        padding="same",
        data_format="channels_last",
        dilation_rate=1,
        groups=1,
        use_skip_connection=False,
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=1,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            use_skip_connection=use_skip_connection,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer, 
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Models.AutoEncoder", name="FCN2D")
class FCN2D(BaseFCN):
    """
    2-D implementation of Fully Convolutional Network (FCN)

    The model is build from a convolutional encoder- and a decoder path,
    and is able to extract features on different scale of the data
    by bisecting the features with each level of the model.

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    use_skip_connection : bool | list, optional
        Whether to use skip connections (basically transform FCN to a Unet).
        Can be defined per layer, i.e., 
        `use_skip_connection=[True, False, False]` results in a skip connection on the highest level only.
        Defaults to `False`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the (optional) information from skip connections.
        For this parameter to have an impact, there has to be at least one skip connection in the model.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"FCN2D"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        data_format="channels_last",
        dilation_rate=1,
        groups=1,
        use_skip_connection=False,
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=2,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            use_skip_connection=use_skip_connection,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer, 
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Models.AutoEncoder", name="FCN3D")
class FCN3D(BaseFCN):
    """
    3-D implementation of Fully Convolutional Network (FCN)

    The model is build from a convolutional encoder- and a decoder path,
    and is able to extract features on different scale of the data
    by bisecting the features with each level of the model.

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    use_skip_connection : bool | list, optional
        Whether to use skip connections (basically transform FCN to a Unet).
        Can be defined per layer, i.e., 
        `use_skip_connection=[True, False, False]` results in a skip connection on the highest level only.
        Defaults to `False`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the (optional) information from skip connections.
        For this parameter to have an impact, there has to be at least one skip connection in the model.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"FCN3D"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        data_format="channels_last",
        dilation_rate=1,
        groups=1,
        use_skip_connection=False,
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=3,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            use_skip_connection=use_skip_connection,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer, 
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )
    


from keras import Layer
from keras import initializers
from keras import regularizers
from keras import constraints
from keras import ops
from keras.src.layers.input_spec import InputSpec
from keras.src.backend.config import backend
from keras.src.backend import standardize_data_format
from keras.src.utils.argument_validation import standardize_tuple
from importlib import import_module
from functools import partial


class BaseSpectralConv(Layer):
    """
    https://arxiv.org/abs/2010.08895
    
    """
    def __init__(
        self,
        rank,
        filters,
        modes,
        data_format="channels_last",
        use_bias=True,
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_constraint=None,
        bias_constraint=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        name=None,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.rank = rank
        self.filters = filters
        self.modes = standardize_tuple(modes, rank, name="modes")
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.data_format = standardize_data_format(data_format)

        fft_module = import_module(name="....ops.fft", package=__package__)
        self.rfft_fn = getattr(fft_module, "rfft" if self.rank == 1 else f"rfft{self.rank}")
        self.irfft_fn = getattr(fft_module, "irfft" if self.rank == 1 else f"irfft{self.rank}")

        # checks
        if self.filters is not None and self.filters <= 0:
            raise ValueError(
                "Invalid value for argument `filters`. Expected a strictly "
                f"positive value. Received filters={self.filters}."
            )
        
        if not all(self.modes):
            raise ValueError(
                "The argument `modes` cannot contain 0. Received "
                f"modes={self.modes}."
            )
        
    def build(self, input_shape):
        if self.built:
            return

        # get data axes
        axes = list(range(len(input_shape)))

        if self.data_format == "channels_last":
            channel_axis = -1
            input_channel = input_shape[-1]

            # if data format is `"channels_last"`, we have to transpose in order to apply the rfft and irfft along the last axes
            transpose_axes = axes.copy()
            inverse_transpose_axes = axes.copy()

            transpose_axes.insert(1, transpose_axes.pop(-1))
            inverse_transpose_axes.append(inverse_transpose_axes.pop(1))

            self.transpose = partial(ops.transpose, axes=transpose_axes)
            self.inverse_transpose = partial(ops.transpose, axes=inverse_transpose_axes)

        else:
            # NOTE this does not matter too much since currently the layer is restricted to use `"channels_last"` data format!
            channel_axis = 1
            input_channel = input_shape[1]

            # if data format is already `"channels_first"`, we do not have to transpose in order to apply the rfft and irfft
            self.transpose = lambda x: x
            self.inverse_transpose = lambda x: x

        # check pad with
        self.pad_width = (
                (0, 0),
                (0, 0), 
                *[(0, s // 2 + 1 - m if i == (len(self.modes) - 1) else s - m) for i, (m, s) in enumerate(zip(self.modes, input_shape[(1 if self.data_format == "channels_last" else 2):]))]
            )
        if list(filter(lambda x: x < (0, 0), self.pad_width)):
                raise ValueError("Too many modes for input shape!")
        
        self.input_spec = InputSpec(
            min_ndim=self.rank + 2, axes={channel_axis: input_channel}
        )

        kernel_shape = (input_channel, self.filters, *self.modes)

        # define real- and imaginary weights
        self._real_kernel = self.add_weight(
            name="real_kernel",
            shape=kernel_shape,
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            trainable=True,
            dtype=self.dtype
        )
        self._imag_kernel = self.add_weight(
            name="imag_kernel",
            shape=kernel_shape,
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            trainable=True,
            dtype=self.dtype
        )

        if self.use_bias:
            self._real_bias = self.add_weight(
                name="bias",
                shape=(self.filters,), 
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                trainable=True,
                dtype=self.dtype
            )
            self._imag_bias = self.add_weight(
                name="bias",
                shape=(self.filters,), 
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                trainable=True,
                dtype=self.dtype
            )
        else:
            self._real_bias = None
            self._imag_bias = None

        """
        The layer operates in the complex Fourier space.
        Here, it is always `data_format="channels_first"`, such that the RFFT can be applied along the last axis or axes.

        Now, we need a slice object to truncate the mode / reduce the complex data to the relevant modes.
        The first two dimensions are the batch and the channel/filters. Then come the modes.

        In 1-D, the RFFT output is just `[0, *positive_freqs]`, shape is `(batch, channels, n // 2 + 1)`.

        In 2-D, we have a 2-D output with shape `(batch, channels, n, n // 2 + 1)`,
        because we have only positive frequencies in `x` but the full spectrum in `y`.

        we can apply an fftshift along the first feature dimension, which would make things way easier!
        In `y`-direction, we will then have `[*negative_freqs, 0, *positive_freqs]`.

        Note that the modes have to be doubled in order to see `m` positive and negative modes!

        """
        feature_axes = axes.copy()
        feature_axes.pop(0)  # remove batch dimension
        feature_axes.pop(channel_axis)  # -1 if self.data_format == "channels_last" else 1)  # remove channel dimension

        self.feature_dims = tuple(input_shape[a] for a in feature_axes)

        # derive rfft shift arguments using `ops.roll`
        # shift have to be (m // 2 for m in modes) for all modes but the last
        # axis have to be all feature axes except for the last!
        _truncation_shifts = [m // 2 for m in self.modes]
        _truncation_shifts.pop(-1)  # remove last since there is no shift in RFFTN along the last axis
        self._truncation_shifts = tuple(_truncation_shifts)

        shift_axes = feature_axes.copy()
        shift_axes.pop(-1)  # remove last axis
        if self.data_format == "channels_last":
            shift_axes = [a + 1 for a in shift_axes]
        self.shift_axes = tuple(shift_axes)
        
        self.mode_truncation_slice = tuple([slice(None), slice(None), *[slice(None, m) for m in self.modes]])

        # declare einsum operation to apply weights
        einsum_dim = "".join([d for _, d in zip(self.modes, ["X", "Y", "Z"])])  # einsum dimensions are just letters for each mode, i.e., "XY" for modes=(8, 16)
        self.einsum_op_forward = f"bi{einsum_dim},io{einsum_dim}->bo{einsum_dim}"
        self.einsum_op_bias = f"bo{einsum_dim},o->bo{einsum_dim}"

        if backend() == "tensorflow":
            # Backpropagation with `tensorflow` backend is a bit cumbersome and requires the exact gradient flow.
            # Therefore, we must declare additional `einsum_ops`.
            self.einsum_op_backprop_weights = f"bo{einsum_dim},bi{einsum_dim}->io{einsum_dim}"
            self.einsum_op_backprop_x = f"bo{einsum_dim},io{einsum_dim}->bi{einsum_dim}"
            self.einsum_op_backprop_bias = f"bo{einsum_dim}->o"  # sum over all axis except output channels

        self.built = True

    def rfft(self, x):
        """
        Performs fast Fourier transform on the real-valued `inputs`.
        
        Parameters
        ----------
        inputs : KerasTensor
            Real-valued input tensor.
        
        Returns
        -------
        (y_real, y_imag) : (KerasTensor, KerasTensor)
            Real- and imaginary part of fast Fourier tranform applied to `inputs`.

        Notes
        -----
        `inputs` must be of shape `(batch, channels, *features)`.

        """
        
        x = self.transpose(x)
        x_real, x_imag = self.rfft_fn(x)

        # # scale outputs for numerical stability in Fourier space
        # x_real /= self.rfft_scaling
        # x_imag /= self.rfft_scaling

        return x_real, x_imag
    
    def irfft(self, inputs):
        """
        Performs inverse fast Fourier transform on the `inputs`.
        
        Parameters
        ----------
        inputs : tuple
            Tuple of `KerasTensor` (real and imaginary part).
        
        Returns
        -------
        outputs : KerasTensor
            Real-valued output of inverse fast Fourier transform applied to `inputs`.

        Notes
        -----
        `inputs` must be of shape `((batch, channels, *features), (batch, channels, *features))`.

        """

        x_real, x_imag = inputs
        y_real = self.irfft_fn((x_real, x_imag))

        # # scale back to "normal" scale
        # y_real *= self.rfft_scaling

        return self.inverse_transpose(y_real)

    def truncation_shift(self, inputs, inverse=False):
        """
        Shifts the Fourier transformed data about `modes` in all directions except `x`.

        Parameters
        ----------
        inputs : KerasTensor
            Real-valued input tensor.
        inverse : bool, optional
            Whether to invert the shift that was applied by calling `truncation_shift` once.
            Defaults to `False`.

        shifted_inputs : KerasTensor
            Shifted version of `inputs`

        """

        if self.rank == 1:
            # if `self.rank==1`, we do not have to shift the inputs!
            return inputs
        
        truncation_shifts = tuple([-s for s in self._truncation_shifts]) if inverse else self._truncation_shifts

        return ops.roll(inputs, shift=truncation_shifts, axis=self.shift_axes)
    
    def call(self, inputs):
        """
        Forward (and backprop) of BaseSpectralConv layer

        The layer first applies a RFFT, truncates the data such that only `self.modes` remain,
        applies the weights, pads the truncated data to match its initial shape.
        The padded data is then transformed using an IRFFT call.

        Since the RFFT and IRFFT are applied along the last axes by default,
        the `inputs` are always transposed to `data_format="channels_first"`,
        and the `outputs` are eventually transformed back to the initial `data_format`.

        Tensorflow can now handle the RFFT and IRFFT calls.
        Hence, the backprop has to be explicitly defined here.
        Thus, when using `backend()="tensorflow"`, this function returns both,
        the output of the spectral convolution `y` and the gradient function `grad`.
        
        Parameters
        ----------
        inputs : KerasTensor
            Input to `SpectralConv1D` layer.

        Returns
        -------
        y | (y, grad) : KerasTensor | (KerasTensor, callable)
            If Tensorflow backend is used, the function returns a tuple of the output of spectral convolution `y` and the gradient `grad`.
            If JAX backend is used, this function returns only the output of the spectral conv `y`

        Notes
        -----
        Since Keras does not have native support for complex dtypes,
        the real- and imaginary parts are handled as two real-valued tensors of `self.dtype`.
        Hence, there are two real-valued weights, `self._real_kernel` and `self._imag_kernel`,
        which are applied to the respective inputs.
        The bias is shared among real- and imaginary parts.

        """

        if backend() == "tensorflow":
            """

            Parameters
            ----------
            inputs : KerasTensor
                Input to `SpectralConv1D` layer.

            Returns
            -------
            (y, grad) : (KerasTensor, callable)
                Tuple of the output of `SpectralConv1D` and the gradient.

            """

            @ops.custom_gradient
            def forward(inputs):
                """
                Custom gradient for `tensorflow` backend
            
                Parameters
                ----------
                inputs : KerasTensor
                    Input to `SpectralConv1D` layer.

                Returns
                -------
                (y, grad) : (KerasTensor, callable)
                    Tuple of the output of `SpectralConv1D` and the gradient.
                    
                """
                
                # forward pass, input shape = (None, *x, ch_in)
                x_real, x_imag = self.rfft(inputs)  # (None, ch_in, *x)

                # apply fft shift
                x_real = self.truncation_shift(x_real)  # (None, ch_in, *x)
                x_imag = self.truncation_shift(x_imag)  # (None, ch_in, *x)

                # reduce to relevant modes
                x_real_truncated = x_real[self.mode_truncation_slice]  # (None, ch_in, *m)
                x_imag_truncated = x_imag[self.mode_truncation_slice]  # (None, ch_in, *m)

                # apply weights
                y_real_truncated = ops.einsum(self.einsum_op_forward, x_real_truncated, self._real_kernel) - ops.einsum(self.einsum_op_forward, x_imag_truncated, self._imag_kernel)  # (None, ch_out, *m)
                y_imag_truncated = ops.einsum(self.einsum_op_forward, x_imag_truncated, self._real_kernel) + ops.einsum(self.einsum_op_forward, x_real_truncated, self._imag_kernel)  # (None, ch_out, *m)

                # add bias
                if self.use_bias:
                    y_real_truncated = ops.einsum(self.einsum_op_bias, y_real_truncated, self._real_bias)  # (None, ch_out, *m)
                    y_imag_truncated = ops.einsum(self.einsum_op_bias, y_imag_truncated, self._imag_bias)  # (None, ch_out, *m)

                # pad to initial size
                y_real = ops.pad(y_real_truncated, pad_width=self.pad_width)  # (None, ch_out, *x)
                y_imag = ops.pad(y_imag_truncated, pad_width=self.pad_width)  # (None, ch_out, *x)

                # apply ifft shift
                y_real = self.truncation_shift(y_real, inverse=True)  # (None, ch_out, *x)
                y_imag = self.truncation_shift(y_imag, inverse=True)  # (None, ch_out, *x)

                # reconstruct y via irfft
                y = self.irfft((y_real, y_imag))  # (None, *x, ch_out)

                def backprop(dy, variables=None):
                    """
                    Backpropagation through the `SpectralConv1D` layer

                    Parameters
                    ----------
                    dy : KerasTensor
                        Gradient of `y`.
                    variables : list, optional
                        List of variables.
                        Defaults to `None`

                    Returns
                    -------
                    (dx, dw) : (KerasTensor, list)
                        Tuple of the gradient of `x` and a list containing the gradients of the weights.
                    
                    """

                    # input shape (None, *x, ch_out)
                    # get real and imaginary part via rfft
                    dy_real, dy_imag = self.rfft(dy)  # (None, ch_out, *x)
                    
                    # apply fft shift
                    dy_real = self.truncation_shift(dy_real)  # (None, ch_out, *x)
                    dy_imag = self.truncation_shift(dy_imag)  # (None, ch_out, *x)

                    # reduce to relevant modes
                    dy_real_truncated = dy_real[self.mode_truncation_slice]  # (None, ch_out, *m)
                    dy_imag_truncated = dy_imag[self.mode_truncation_slice]  # (None, ch_out, *m)

                    # compute gradients for weights
                    dw_real = ops.einsum(self.einsum_op_backprop_weights, dy_real_truncated, x_real_truncated) + ops.einsum(self.einsum_op_backprop_weights, dy_imag_truncated, x_imag_truncated)  # (None, ch_out, *m)
                    dw_imag = ops.einsum(self.einsum_op_backprop_weights, dy_imag_truncated, x_real_truncated) - ops.einsum(self.einsum_op_backprop_weights, dy_real_truncated, x_imag_truncated)  # (None, ch_out, *m)

                    if self.use_bias:
                        # compute gradients for bias
                        db_real = ops.einsum(self.einsum_op_backprop_bias, dy_real_truncated)  # (None, ch_out, *m)
                        db_imag = ops.einsum(self.einsum_op_backprop_bias, dy_imag_truncated)  # (None, ch_out, *m)

                    # compute gradient for inputs
                    dx_real_truncated = ops.einsum(self.einsum_op_backprop_x, dy_real_truncated, self._real_kernel) + ops.einsum(self.einsum_op_backprop_x, dy_imag_truncated, self._imag_kernel)  # (None, ch_in, *m)
                    dx_imag_truncated = ops.einsum(self.einsum_op_backprop_x, dy_imag_truncated, self._real_kernel) - ops.einsum(self.einsum_op_backprop_x, dy_real_truncated, self._imag_kernel)  # (None, ch_in, *m)

                    # pad to initial size
                    dx_real = ops.pad(dx_real_truncated, pad_width=self.pad_width)  # (None, ch_in, *x)
                    dx_imag = ops.pad(dx_imag_truncated, pad_width=self.pad_width)  # (None, ch_in, *x)

                    # apply ifft shift
                    dx_real = self.truncation_shift(dx_real, inverse=True)  # (None, ch_in, *x)
                    dx_imag = self.truncation_shift(dx_imag, inverse=True)  # (None, ch_in, *x)

                    # apply irfft
                    dx = self.irfft((dx_real, dx_imag))  # (None, *x, ch_in)
                    
                    if self.use_bias:
                        grads = [db_real, db_imag, dw_real, dw_imag]
                    else:
                        grads = [dw_real, dw_imag]
                    
                    return dx, grads

                return y, backprop
                
            return forward(inputs)

        if backend() == "jax":
            """
            
            Parameters
            ----------
            inputs : KerasTensor
                Input to `SpectralConv1D` layer.

            Returns
            -------
            y : KerasTensor
                The output of `SpectralConv1D`.

            """

            # forward pass, input shape = (None, *x, ch_in)
            x_real, x_imag = self.rfft(inputs)  # (None, ch_in, *x)

            # apply fft shift
            x_real = self.truncation_shift(x_real)  # (None, ch_in, *x)
            x_imag = self.truncation_shift(x_imag)  # (None, ch_in, *x)

            # reduce to relevant modes
            x_real_truncated = x_real[self.mode_truncation_slice]  # (None, ch_in, *m)
            x_imag_truncated = x_imag[self.mode_truncation_slice]  # (None, ch_in, *m)

            # apply weights
            y_real_truncated = ops.einsum(self.einsum_op_forward, x_real_truncated, self._real_kernel) - ops.einsum(self.einsum_op_forward, x_imag_truncated, self._imag_kernel)  # (None, ch_out, *m)
            y_imag_truncated = ops.einsum(self.einsum_op_forward, x_imag_truncated, self._real_kernel) + ops.einsum(self.einsum_op_forward, x_real_truncated, self._imag_kernel)  # (None, ch_out, *m)

            # add bias
            if self.use_bias:
                y_real_truncated = ops.einsum(self.einsum_op_bias, y_real_truncated, self._real_bias)  # (None, ch_out, *m)
                y_imag_truncated = ops.einsum(self.einsum_op_bias, y_imag_truncated, self._imag_bias)  # (None, ch_out, *m)

            # pad to initial size
            y_real = ops.pad(y_real_truncated, pad_width=self.pad_width)  # (None, ch_out, *x)
            y_imag = ops.pad(y_imag_truncated, pad_width=self.pad_width)  # (None, ch_out, *x)

            # apply ifft shift
            y_real = self.truncation_shift(y_real, inverse=True)  # (None, ch_out, *x)
            y_imag = self.truncation_shift(y_imag, inverse=True)  # (None, ch_out, *x)

            # reconstruct y via irfft
            y = self.irfft((y_real, y_imag))  # (None, *x, ch_out)

            return y

        raise NotImplementedError(f"The call method is only implemented for keras backends `'tensorflow'` and `'jax'`")

    def compute_output_shape(self, input_shape):
        """
        Compute output shape of `BaseSpectralConv`

        Parameters
        ----------
        input_shape : tuple
            Input shape.

        Returns
        -------
        output_shape : tuple
            Output shape.

        """

        input_shape: list = list(input_shape)
        channel_axis = -1 if self.data_format == 'channels_last' else 1

        input_shape[channel_axis] = self.filters
        return tuple(input_shape)

    def get_config(self):
        """
        Get config method.
        Required for serialization.

        Returns
        -------
        config : dict
            Dictionary with the configuration of `BaseSpectralConv`.

        Notes
        -----
        The `config` does not contain the `self.rank` parameter,
        which is not required when the class is subclassed with hard-coded `rank`.

        """
        
        config = super().get_config()
        config.update({
            "filters": self.filters,
            "modes": self.modes,
            "data_format": self.data_format,
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint),
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer)
        })
        return config

    @classmethod
    def from_config(cls, config):
        """
        Necessary for Keras deserialization

        Parameters
        ----------
        cls : BasBaseSpectralConvFCN
            The `BaseSpectralConv` class.
        config : dict
            Dictionary with the layer configuration.

        Returns
        -------
        cls : BaseSpectralConv
            Instance of `BaseSpectralConv` from `config`.
            
        """

        kernel_initializer_cfg = config.pop("kernel_initialzer")
        bias_initializer_cfg = config.pop("bias_initialzer")
        kernel_constraint_cfg = config.pop("kernel_constraint")
        bias_constraint_cfg = config.pop("bias_constraint")
        kernel_regularizer_cfg = config.pop("kernel_regularizer")
        bias_regularizer_cfg = config.pop("bias_regularizer")

        config.update({
            "kernel_initializer": initializers.deserialize(kernel_initializer_cfg),
            "bias_initializer": initializers.deserialize(bias_initializer_cfg),
            "kernel_constraint": constraints.deserialize(kernel_constraint_cfg),
            "bias_constraint": constraints.deserialize(bias_constraint_cfg),
            "kernel_regularizer": regularizers.deserialize(kernel_regularizer_cfg),
            "bias_regularizer": regularizers.deserialize(bias_regularizer_cfg)
        })

        return cls(**config)



from .attention import SpatialAttention, ChannelAttention
from .cbam import CBAM
from .film import FiLM
from .residual import Residual
from .temporal_slice import TemporalSlice
from .weight_normalization import WeightNormalization


from keras import saving
from .base_models import BaseUnet, BaseSmoothUnet


@saving.register_keras_serializable(package="Kerex.Models.AutoEncoder", name="Unet1D")
class Unet1D(BaseUnet):
    """
    1-D implementation of Unet, cf. [Ronneberger et al.](https://arxiv.org/abs/1505.04597)

    The model is build from a convolutional encoder- and a decoder path,
    and is able to extract features on different scale of the data
    by bisecting the features with each level of the model.
    Additionally, skip connections feed through high-resolution features from the encoder 
    to the decoder on each level.

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    padding : str, optional {`"same"`, `"causal"`}
        Padding for all convolutional layers in the model.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the information from skip connections.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"Unet1D"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        padding="same",
        data_format="channels_last",
        dilation_rate=1,
        groups=1,
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=1,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer, 
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Models.AutoEncoder", name="Unet2D")
class Unet2D(BaseUnet):
    """
    2-D implementation of Unet, cf. [Ronneberger et al.](https://arxiv.org/abs/1505.04597)

    The model is build from a convolutional encoder- and a decoder path,
    and is able to extract features on different scale of the data
    by bisecting the features with each level of the model.
    Additionally, skip connections feed through high-resolution features from the encoder 
    to the decoder on each level.

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the information from skip connections.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"Unet2D"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        filters=[8, 16, 32],
        kernel_size=5,
        data_format="channels_last",
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=2,
            filters=filters,
            kernel_size=kernel_size,
            data_format=data_format,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer, 
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Models.AutoEncoder", name="Unet3D")
class Unet3D(BaseUnet):
    """
    3-D implementation of Unet, cf. [Ronneberger et al.](https://arxiv.org/abs/1505.04597)

    The model is build from a convolutional encoder- and a decoder path,
    and is able to extract features on different scale of the data
    by bisecting the features with each level of the model.
    Additionally, skip connections feed through high-resolution features from the encoder 
    to the decoder on each level.

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the information from skip connections.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"Unet3D"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        data_format="channels_last",
        dilation_rate=1,
        groups=1,
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=3,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer, 
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


### MODELS THAT USE SMOOTH UPSAMPLING LAYER ###
@saving.register_keras_serializable(package="Kerex.Models.AutoEncoder", name="SmoothUnet1D")
class SmoothUnet1D(BaseSmoothUnet):
    """
    1-D implementation of Unet, cf. [Ronneberger et al.](https://arxiv.org/abs/1505.04597)

    The model is build from a convolutional encoder- and a decoder path,
    and is able to extract features on different scale of the data
    by bisecting the features with each level of the model.
    Additionally, skip connections feed through high-resolution features from the encoder 
    to the decoder on each level.
    Upsampling is performed via a stack of image upsampling and a convolution.

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    padding : str, optional {`"same"`, `"causal"`}
        Padding for all convolutional layers in the model.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the information from skip connections.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"Unet1D"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        padding="same",
        data_format="channels_last",
        dilation_rate=1,
        groups=1,
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=1,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer, 
            interpolation=None,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Models.AutoEncoder", name="SmoothUnet2D")
class SmoothUnet2D(BaseSmoothUnet):
    """
    2-D implementation of Unet, cf. [Ronneberger et al.](https://arxiv.org/abs/1505.04597)

    The model is build from a convolutional encoder- and a decoder path,
    and is able to extract features on different scale of the data
    by bisecting the features with each level of the model.
    Additionally, skip connections feed through high-resolution features from the encoder 
    to the decoder on each level.
    Upsampling is performed via a stack of image upsampling and a convolution.

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the information from skip connections.
        Defaults to `"concatenate"`.
    interpolation : str, optional {`"nearest"`, `"bicubic"`, `"bilinear"`, `"lanczos3"`, `"lanczos5"`}
        Interpolation to use in `SmoothUpSampling2D` layer.
        This option is only valid for `rank=2` and is ignored otherwise.
        Defaults to `"nearest"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"Unet2D"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        filters=[8, 16, 32],
        kernel_size=5,
        data_format="channels_last",
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        interpolation="nearest",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        assert interpolation in ["bicubic", "bilinear", "lanczos3", "lanczos5", "nearest"], f"Unsupported interpolation '{interpolation}'. Supported interpolations are ['bicubic', 'bilinear', 'lanczos3', lanczos5', 'nearest']"
        super().__init__(
            rank=2,
            filters=filters,
            kernel_size=kernel_size,
            data_format=data_format,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer, 
            interpolation=interpolation,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Models.AutoEncoder", name="SmooothUnet3D")
class SmoothUnet3D(BaseSmoothUnet):
    """
    3-D implementation of Unet, cf. [Ronneberger et al.](https://arxiv.org/abs/1505.04597)

    The model is build from a convolutional encoder- and a decoder path,
    and is able to extract features on different scale of the data
    by bisecting the features with each level of the model.
    Additionally, skip connections feed through high-resolution features from the encoder 
    to the decoder on each level.
    Upsampling is performed via a stack of image upsampling and a convolution.

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the information from skip connections.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"Unet3D"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        data_format="channels_last",
        dilation_rate=1,
        groups=1,
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=3,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer, 
            interpolation=None,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )